/*
 * All ICMP time-stamp processing routines
 * Santosh Krishnan
 * 03/01
 */

#include "def.h"

int                    icp_fd;         /* ICMP socket fd              */

static int             icp_seq;        /* Sequence number in ICMP     */
static int             icp_ident;      /* Identifier to receive       */
static struct sockaddr_in dest;        /* Peer of ICMP sender         */
static int             icmp_seq_range; /* Range equals batch file size*/
static int             icmp_payload_len; /* Length of ICMP payload    */

u_char                 sendpack[PACKSIZE]; /* ICMP sending packet      */
u_char                 recvpack[PACKSIZE]; /* Received is IP, not ICMP */ 



/*
 * Initialize the sockets for ICMP
 * Called by: probe.c:: start_probes
 */
icmp_init ()
{
  struct protoent    *proto;

  if ((proto = getprotobyname ("icmp")) == NULL)
    perror ("Unknown protocol: icmp");
  if ((icp_fd = socket (AF_INET, SOCK_RAW, proto->p_proto)) < 0)
    perror ("Can't create raw socket");

  icp_seq = 0;
  /* icmp_seq/Total Hops gives the index into the Batch File
     while icmp_seq % Total_Hops gives the hop number */
  icp_ident = getpid () & 0xffff;
  icmp_seq_range = Total_Hops * Total_Probes;
  icmp_payload_len = ICMPSIZE;
#ifdef ICMPMICROSEC
  icmp_payload_len += sizeof (struct timeval);
#endif

  /* Initialize dest */
  bzero ((char *) &dest, sizeof (dest));
  dest.sin_family = AF_INET;

  return;
}



/*
 * Send an ICMP packet to each of the hosts
 * Schedule the next time the packets are sent 
 * Called by: probe.c:: probe_loop
 *            Alarm by self
 */
void
icmp_send (int dummy)
{
  register int            i, j;
  register struct icmp    *icp;
  struct timeval tv;

  alarm (Probe_Gap);   /* Schedule next bunch of ICMP packets */

  /* Send the ICMP packets */
  probe_count ++;
#ifdef DEBUG_TIMING
  printf ("Sending ICMP packet %d\n", probe_count);
#endif

  for (j=0; j<Total_Hops; j++) {
    icp = (struct icmp *) sendpack;
    icp->icmp_type = ICMP_TSTAMP;
    icp->icmp_code = 0;
    icp->icmp_cksum = 0;
    icp->icmp_id = icp_ident;
    icp->icmp_seq = icp_seq++;

    /* We dont care about wrap-arounds */
    if (icp_seq == icmp_seq_range) icp_seq = 0;
    
    /* Timing */
    gettimeofday ((struct timeval *) &tv, (struct timezone *) 0); 
    icp->icmp_otime = (u_long) (((tv.tv_sec % 86400) * 1000)
				+ (tv.tv_usec/1000));
    icp->icmp_rtime = (u_long) 0;
    icp->icmp_ttime = (u_long) 0;
#ifdef ICMPMICROSEC
    /* Send timeval in the packet because ICMP timestamps
       use milliseconds granularity
       */
    bcopy ((char *) &tv, (char *) &sendpack[ICMPSIZE], sizeof(struct timeval));
#endif
    icp->icmp_cksum = in_cksum (icp, icmp_payload_len);
    
    /* Send the ICMP packet */
    dest.sin_addr.s_addr = inet_addr (Hop_List[j].hop_addr);
    
    if ((i = sendto (icp_fd, sendpack, icmp_payload_len, 0,
		     (struct sockaddr *) &dest, sizeof (dest))) != 
	icmp_payload_len) {
      if (i < 0)
	perror ("ICMP sendto error");
      else
	perror ("ICMP didnt send entire packet");
    }
  }
  return;
}



/*
 * Assemble the received ICMP info
 * into the batch file
 * Called by: meas.c:: meas_loop
 * Calls: icmp-proc.c:: process_recv
 */
icmp_recv ()
{
  register int       recv_size;
  int                fromlen;
  struct sockaddr_in from;
  struct timeval     tv;

  fromlen = sizeof (from);
  if ((recv_size = recvfrom (icp_fd, recvpack, sizeof (recvpack), 0,
			     (struct sockaddr *) &from, &fromlen)) < 0) {
    if (errno == EINTR)
      return;
    perror ("ICMP receive error");
    return;
  }
  /* Timestamp reply */
  gettimeofday ((struct timeval *) &tv, (struct timezone *) 0);     
  process_recv (recv_size, &from, &tv); 
  return;
}



/*
 * Process the received ICMP packets 
 * Called by: icmp_recv
 * Calls meas.c:: add_to_batch_file
 */
process_recv (int recv_size, struct sockaddr_in *from, 
	      struct timeval *tv)
{
  register int          hop, seq;
  register int          iphdrlen;
  struct ip             *ip;      /* ptr to IP header   */
  struct icmp           *icp;     /* ptr to ICMp header */
  u_long                timestamp;
  
  timestamp =  (u_long) (((tv->tv_sec % 86400) * 1000)
			 + (tv->tv_usec/1000));
  ip = (struct ip *) recvpack;
  iphdrlen = ip->ip_hl << 2;
  if (recv_size < (iphdrlen + icmp_payload_len))
    /* Packet too short for our consideration */
    return;
  recv_size -= iphdrlen;

  icp = (struct icmp *) (recvpack + iphdrlen);
  if (icp->icmp_type != ICMP_TSTAMPREPLY)
    /* Not a timestamp response packet */
    return;

  if (icp->icmp_id != icp_ident)
    /* Not our packet */
    return;
  
  /* Check integrity and send to
     the measurement process to
     add into the batch file */
  seq = icp->icmp_seq;
  hop = seq % Total_Hops;  /* Hop index   */
  seq = seq / Total_Hops;  /* Probe index */
#ifdef DEBUG_ICMP
  if (ntohl (from->sin_addr.s_addr) != inet_addr (Hop_List[hop].hop_addr))
    printf ("Warning: Address in hop list does not match\n");
#endif
#ifdef ICMPMICROSEC
  tvsub (tv, (struct timeval *) &recvpack[iphdrlen + ICMPSIZE]);
  add_to_batch_file (hop, seq, 0,
		     icp->icmp_rtime * 1000,
		     icp->icmp_ttime * 1000,
		     tv->tv_sec * 1000000 + tv->tv_usec);
#else
  add_to_batch_file (hop, seq, icp->icmp_otime,
		     icp->icmp_rtime, icp->icmp_ttime,
		     timestamp);
#endif
  return;
}


/*
 * Subtract timevals
 * Called by: process_recv
 */
tvsub (struct timeval *out, struct timeval *in)
{
  /* subtract microseconds */
  if ((out->tv_usec -= in->tv_usec) < 0) {
    out->tv_sec-- ;                       
    out->tv_usec += 1000000; 
  }
  /* subtract seconds */
  out->tv_sec -= in->tv_sec;
  return;
}


/*
 * Compute checksum
 * Called by icmp-proc.c:: icmp_send
 */
in_cksum (u_short *ptr, int nbytes)
{
  register long    sum;
  u_short          oddbyte;
  register u_short answer;

  /* Borrowed from Steven's pg. 454-455 */

  sum = 0;
  while (nbytes > 1) {
    sum += *ptr++;
    nbytes -= 2;
  }
  if (nbytes == 1) {
    oddbyte = 0;
    *((u_char *) &oddbyte) = *(u_char *) ptr;
    sum += oddbyte;
  }

  sum = (sum >> 16) + (sum & 0xffff);
  sum += (sum >> 16);
  answer = ~sum;
  return (answer);
}
