/*
 * Per Hop Internet Traffic Measurement Tool
 * Main program: Initiates batches, calls traceroute
 *
 * Santosh Krishnan 03/01
 */

#include "def.h"

/* Hop List data structure */
hop_entry              Hop_List[MAXHOPS];
int                    Total_Hops;




/*
 * Main program
 * Takes in the parameters and calls functions
 * to start probing the hops
 *
 * Calls main.c:: init_hops
 *       probe.c:: start_probes
 */
main (int argc, char *argv[])
{
  /* Check command line */
  if ((argc < 2) || ((argc > 2) && (argc != 4))) {
    printf ("Usage: tmtool <host name/ip address> [N T]\n");
    printf ("       N: number of probes per batch > %d\n", MINPROBES);
    printf ("       T: Time in seconds between probes > %d (sec)\n",
	    MINPROBEGAP);
    exit (1);
  }
  else if (argc == 4) {
    /* Accept the number of probes per batch and 
       the probe sample rate
       */
    if ((Num_Probes = atoi (argv[2])) < MINPROBES) {
      Num_Probes = MINPROBES;
      printf ("Warning: Number of probes set to minimum accepted: %d\n",
	      Num_Probes);
    }
    if ((Probe_Gap = atoi (argv[3])) < MINPROBEGAP) {
      Probe_Gap = MINPROBEGAP;
      printf ("Warning: Probe sampling set to highest possible period: %d\n",
	      Probe_Gap);
    }
  }
    
  printf ("\n%s\n\n", srcid);
  /* Initialize the list of nodes by calling traceroute */
  if (init_hops (argv[1]))
    start_probes (argc, argv);

  exit (0);
}




/*
 * Initializes hop list by calling trace route
 *
 * Called by main.c::main
 * Calls main.c:: parse_addr, init_trace, trace_build
 */
init_hops (char *dest)
{
  char   dest_addr[ADDRSIZE];   /* IP address of remote host */
  char   *src;                  /* My name, return of getenv */
  char   src_addr[ADDRSIZE];

  /* Find if IP address of remote host is valid */
  if (! (parse_addr (dest, dest_addr))) {
    printf ("Remote host (%s) not found. Exiting....\n", dest);
    return (0);
  }
  else {
    /* Find the IP address of sending host */
    src = getenv ("HOSTNAME");
    if (! (parse_addr (src, src_addr))) {
      printf ("Could not resolve self IP address. Exiting....\n");
      return (0);
    }
    else {
      
      /* Initiate trace route to build the hop list of IP addresses */

      if (! (strncmp (src_addr, dest_addr, strlen (src_addr)))) {
	printf ("Invalid usage: Same address for source and destination\n\n");
	exit (1);
      }

      printf ("Starting traceroute.... \n");
      printf ("Sending host: %s (%s)\n", src, src_addr);
      printf ("Remote host: %s (%s)\n\n", dest, dest_addr);
      
      init_trace (src, src_addr, dest, dest_addr);
      trace_build (src, src_addr);
    }
  }
  return (1);
}



/*
 * Parse the given address
 * If it is an IP address, leave it as it is
 * Else return the IP address of remote host
 * Called by main.c:: init_hops
 */
parse_addr(char *addr, char *ip_addr)
{
  struct  hostent *hp;

  if (inet_addr (addr) == -1){
    /* Assume that addr is a host name, not an address */
    hp = gethostbyname (addr);
    if (hp) {
      struct sockaddr_in tmp;
      bcopy(hp->h_addr, (caddr_t)&(tmp.sin_addr), hp->h_length);
      strcpy (ip_addr, (char *) inet_ntoa (tmp.sin_addr));
      return 1;
    } 
    else {
      /* Cannot be resolved */
      return 0;
    }
  }
  else {
    /* Given address is an IP address */
    strcpy (ip_addr, addr);
    return 1;
  }
}



/*
 * Initialize the IPC for trace route list
 * Child process generates list and streams
 * to parent for parsing
 *
 * Called by::init_hops
 */
init_trace (char *src_name, char *src_addr,
	     char *dest_name, char *dest_addr)
{
  int    fd[2];                 /* For the pipe from traceroute */
  pid_t  pid;                   /* For the fork system call     */
  char   trc[TRCMDSIZE];        /* Traceroute message           */
  sprintf (trc, "%s %s", TRACEROUTE, dest_addr);

  if (pipe (fd) < 0)
    perror ("Pipe error");

  /* Fork, let child do the "trace-route" and stream the
     output to the parent who processes it through the pipe     */

  if ( (pid = fork ()) < 0)
    perror ("Fork error in trace_build");

  else if (pid > 0) {           /* Parent process               */

    close (fd[1]);              /* Parent only reads from pipe  */
    if (dup2 (fd[0], STDIN_FILENO) != STDIN_FILENO) 
      perror ("dup error to stdin of trace_build");
    close (fd[0]);              /* Parent just uses stdin       */
  }

  else {                        /* Child process, run traceroute*/
    close (fd[0]);              /* Parent only writes to pipe   */
    if (dup2 (fd[1], STDOUT_FILENO) != STDOUT_FILENO)
      perror ("dup error to stdout of trace_build");
    close (fd[1]);
    fclose (stderr);
    system (trc);
    exit (0);
  }
  return;
}



/*
 * Parse the trace route info and create
 * a list of nodes
 * Called by main.c:: init_hops
 */
trace_build (char *src, char *src_addr)
{
  char   inbuf[BUFSIZ];
  char   name[NAMESIZE];
  char   addr[ADDRSIZE];
  int    num, i;

  /* First fill in the first entry */
  Total_Hops = 1;
  add_hop (src, src_addr, 0);

  /* Read output of traceroute line by line */
  /* CAUTION: This depends highly on the 
     format of the trace-route output       */
  while (fgets (inbuf, BUFSIZ, stdin) != NULL) {
    if (sscanf (inbuf, "%d %s (%[0-9.]", &num, name, addr) == 3) {
      add_hop (name, addr, num);
      Total_Hops ++;
    }
  }

  /* Print out Hop List */
  printf ("List of Hops:\n");
  for (i=0; i<Total_Hops; i++)
    printf ("%d %s (%s)\n", i, Hop_List[i].hop_name,
	    Hop_List[i].hop_addr);
  printf ("\n");
  return;
}



/*
 * Add a hop to the hop list 
 * Called by main.c::trace_build
 */
add_hop (char *name, char *addr, int num)
{
  strcpy (Hop_List[num].hop_name, name);
  strcpy (Hop_List[num].hop_addr, addr);
  return;
}
