/*
 * Processes i.e. measures info in batch files
 * and displays them.
 *
 * Santosh Krishnan
 * 03/01
 */

#include "def.h"

int                    current_probe=0;          /* Index: first unprocessed */
Result                 Result_File;              /* Data structure of results*/
int                    Batch_Time;               /* In seconds               */

/* Batch file parameters */
int                    Total_Batches=1;          /* Number of current batches*/
int                    Total_Probes=MINPROBES;   /* Total probes batch file  */
Probe                  **Batch_File;             /* Unprocessed probe info   */

/* Display parameters */
Display_Samples        Disp;                     /* Temporary display struct */
int                    batch_count=0;            /* only for debugging       */




/*
 * Initialize the batch data structures
 * Called by probe.c::init_probes
 */
init_batches ()
{
  register int    i,j;

  /* The batch file contains the probe values of the current
     batches, i.e., maximum number of unprocessed batches.
     First index is the hop number, the next index is the
     probe number to that hop */

  Batch_File = (Probe **) malloc (sizeof (Probe *) * Total_Hops);
  for (i=0; i<Total_Hops; i++) {
    Batch_File[i] = (Probe *) malloc (sizeof (Probe) * Total_Probes);
    for (j=0; j<Total_Probes; j++)
      Batch_File[i][j].rcvflag = 0;
  }
  return;
}



/*
 * Initialize the results data structure
 * Called by probe.c::start_probes
 */
init_measures ()
{
  register int i;
  register oqp *tmp;

  /* Prepare results file for the first batch */

  Result_File.num_batches = 0;
  Result_File.l_time = (int *) malloc (sizeof (int) * Total_Hops);
  for (i=0; i<Total_Hops; i++)
    Result_File.l_time[i] = MAXLINKTIME;
  
  Result_File.head_oqp = Result_File.tail_oqp
    = (oqp *) malloc (sizeof (oqp));
  
  /* Initialize the first entry. After a batch is
     processed, the subsequent entries have to be malloc-ed */
  tmp = Result_File.head_oqp;
  tmp->o_time = (int *) malloc (sizeof (int) * Total_Hops);
  tmp->q_time = (int *) malloc (sizeof (int) * Total_Hops);
  tmp->d_time = (int *) malloc (sizeof (int) * Total_Hops);
  tmp->p_loss = (double *) malloc (sizeof (double) * Total_Hops);
  tmp->next = NULL;
  for (i=0; i<Total_Hops; i++) {
    tmp->o_time[i] = 0;
    tmp->q_time[i] = 0;
    tmp->d_time[i] = 0;
    tmp->p_loss[i] = (double) 0;
  }
  /* Before display, the q_time actually contains the average of
     the timestamp differences, because the link
     time is calculated as a minimum over all batches */

  Disp.l_time = (int *) malloc (sizeof (int) * Total_Hops);
  Disp.o_time = (int *) malloc (sizeof (int) * Total_Hops);
  Disp.q_time = (int *) malloc (sizeof (int) * Total_Hops);
  Disp.d_time = (int *) malloc (sizeof (int) * Total_Hops);
  Disp.p_loss = (double *) malloc (sizeof (double) * Total_Hops);

  return;
}



/*
 * Displays the results from the results file
 * It is only during display that the link times
 * are separated from the q_times. Hence, if link times
 * change, you have to re-display previous results 
 * 
 * The argument is the batch number
 * Called by: main.c::main
 */
display_results (int num)
{
  register oqp  *tmp;
  register int     i;

  if (num >= Result_File.num_batches) return (0);/* Batch not yet processed */

  tmp = Result_File.head_oqp;
  /* Look for the correct batch */
  for (i=0; i<num; i++)
    tmp = tmp->next;

  adjust_display (tmp);
  /* Print the results for that batch */
  printf("Batch %d:\n\n", num);

  for (i=0; i<Total_Hops; i++) {
    printf ("(Node %2d): ", i);

    printf ("o: %d   l: %d   q: %d  (d: %d) loss: %8.7lf\n", 
	    Disp.o_time[i],
	    Disp.l_time[i],
	    Disp.q_time[i],
	    Disp.d_time[i],
	    Disp.p_loss[i]);
  }
  printf ("\n");
  return (1);
}



/*
 * Assemble display info for the current batch
 * Called by: display_results
 */
adjust_display (oqp *tmp)
{
  register int i, accum;
  register int ltime, qtime;

  /*
   * NOTE: All the accumulators to separate link from
   * OS and queueing times are off for now -- 03/21
   */
  for (i=0; i<Total_Hops; i++) {
    Disp.o_time[i] = tmp->o_time[i];
    Disp.d_time[i] = tmp->d_time[i];
    Disp.p_loss[i] = tmp->p_loss[i];
  }

  accum = 0;
  Disp.l_time[0] = 0;
  /* De-seggregate link and queue times */
  for (i=1; i<Total_Hops; i++) {
    /* Skip the node 0 */
    ltime = Result_File.l_time[i];
    ltime -= accum;
    ltime /= 2;
    Disp.l_time[i] = ltime;
#ifdef INFERENCE_MODEL
    accum += 2*ltime;
#endif
  }  

  accum = 0;
  /* Queue components */
  qtime = Disp.q_time[0] =
    (tmp->q_time[0] - tmp->o_time[0]);
#ifdef INFERENCE_MODEL
  accum += (qtime);
#endif
  
  for (i=1; i<Total_Hops; i++) {
    /* Not the sending node, so process */
#ifdef INFERENCE_MODEL
    accum += (2 * Disp.l_time[i]);
#else
    accum = 2 * Disp.l_time[i];
#endif
    qtime = Disp.q_time[i] = (tmp->q_time[i] - (accum) -
			      tmp->o_time[i] - tmp->o_time[0]);
#ifdef INFERENCE_MODEL
    accum += (2 * qtime);
#endif
  }
  return;
}



/*
 * Periodically measure results
 * Take the raw data from the batch files and compile
 * the results files. 
 *
 * Called by: probe.c:: start_probes
 * Calls: meas.c:: update_batch
 *        icmp-proc.c:: icmp_recv
 */
meas_loop ()
{
  Batch_Time = Num_Probes * Probe_Gap;
  sigset (SIGALRM, update_batch);
  alarm (Batch_Time + RECVTIMEOUT);  /* Wait for time-out for first batch */
  for (;;) icmp_recv ();
  return;
}



#ifdef GUI_ENABLE
/*
 * Measurement loop with GUI
 * Called by probe.c:: start_probes
 * Calls: meas.c:: update_batch
 *        icmp-proc.c:: icmp-recv
 *        gui.c:: process_gui
 */
gui_meas_loop ()
{
  fd_set rset;
  int max_fd;

  Batch_Time = Num_Probes * Probe_Gap;
  sigset (SIGALRM, update_batch);
  alarm (Batch_Time + RECVTIMEOUT);  /* Wait for time-out for first batch */

  /* Manage receipt of ICMP packets and display */
  if (icp_fd > disp_fd) max_fd = icp_fd;
  else max_fd = disp_fd;
  max_fd ++;

  /* Prepare for select */
  FD_ZERO (&rset);
  FD_SET (icp_fd, &rset);
  FD_SET (disp_fd, &rset);

  for (;;) {
    /* Select from GUI event or ICMP read event */
    select (max_fd, &rset, 0, 0, (struct timeval *) 0);
    
    if (FD_ISSET (icp_fd, &rset))
      /* Process ICMP */
      icmp_recv ();
    if (FD_ISSET (disp_fd, &rset))
      /* Process GUI */
      process_gui ();
    
    /* Re-set the fd_set */
    FD_ZERO (&rset);
    FD_SET (icp_fd, &rset);
    FD_SET (disp_fd, &rset);
  }
  return;
}
#endif



/*
 * Add to the batch file 
 * Called by icmp-proc.c:: process_recv
 */
add_to_batch_file (int hop, int seq, u_long t1,
		   u_long t2, u_long t3, u_long t4)
{
  register Probe *prb;

  prb = *(Batch_File + hop) + seq;
  prb->rcvflag = 1;
  prb->T1 = t1;
  prb->T2 = t2;
  prb->T3 = t3;
  prb->T4 = t4;
  return;
}



/*
 * Update batch. Re-instate alaram
 * Depends on CBR transmission of probes
 * If GUI is used, then display results at end of loop
 * Called by: meas.c:: meas_loop
 * Calls: add_results_node, process_otime, process_ltime, process_qtime
 *        gui.c:: gui_redraw, adjust_gui_display
 */
void
update_batch (int dummy)
{
  alarm (Batch_Time);

  batch_count++;
  /* Process batch */
#ifdef DEBUG_TIMING
  printf ("Batch processed %d\n", batch_count);
#endif

  process_otime ();
  process_ltime ();
  process_qtime ();
  process_dtime ();
  process_loss ();  /* rcvflag is reset here */

  current_probe = (current_probe + Num_Probes) % Total_Probes;
  add_results_node ();
#ifdef GUI_ENABLE
  adjust_gui_display ();
  gui_redraw ();
#endif
  return;
}



/*
 * Calculate loss probability within the batch
 * Called by: meas.c:: update_meas
 */
process_loss ()
{
  register oqp    *tmp;
  register Probe  *prb;

  register int    j,k;
  register int    nback;
  register double loss;
  register double prod_1_loss;

  /* Start with the tail */
  tmp = Result_File.tail_oqp;
  prod_1_loss = (double) 1;

  for (k=0; k<Total_Hops; k++) {
    /* Check the hops in order */
    nback = Num_Probes;
    prb = *(Batch_File + k) + current_probe;

    for (j=0; j<Num_Probes; j++, prb++) {
      /* Look at each entry of the batch file and 
	 compute first order results */
      if (prb->rcvflag) {
	prb->rcvflag = 0;
      }
      else {
	nback --;
	continue;
      }
    }
    /* Calculate loss */
    loss = 1 - (((double) nback / (double) Num_Probes)/
		prod_1_loss);
    prod_1_loss *= (1 - loss);
    tmp->p_loss[k] = loss;
  }
  return;
}



/*
 * Process OS time
 * Called by meas.c:: update_meas
 */
process_otime ()
{
  register oqp    *tmp;
  register Probe  *prb;

  register int    j,k;
  register int    accum;
  register int    nrcvd;

  /* Start with the tail */
  tmp = Result_File.tail_oqp;

  for (k=0; k<Total_Hops; k++) {
    /* Check the hops in order */
    prb = *(Batch_File + k) + current_probe;
    nrcvd = 0;
    accum = 0;

    for (j=0; j<Num_Probes; j++, prb++) {
      /* Look at each entry of the batch file and 
	 compute first order results */
      if (prb->rcvflag) {
	nrcvd ++;
	accum += (prb->T3 - prb->T2); 
      }
    }
    /* Calculate mean */
    if (nrcvd)
      tmp->o_time[k] = (int) ((double) accum / (double) nrcvd);
  }
  return;
}



/*
 * Process link time
 * Called by meas.c:: update_meas
 */
process_ltime ()
{
  register oqp    *tmp;
  register Probe  *prb;
  register int    j,k;
  register int    accum=MAXLINKTIME;

  /* Start with the tail */
  tmp = Result_File.tail_oqp;
  
  for (k=0; k<Total_Hops; k++) {
    /* Check the hops in order */
    prb = *(Batch_File + k) + current_probe;
    
    for (j=0; j<Num_Probes; j++, prb++) {
      /* Just keep track of minimum */
      if (prb->rcvflag)
	accum = (prb->T4 - prb->T1) - (prb->T3 - prb->T2); 
    }
    /* Calculate min */
    if (accum < Result_File.l_time[k])
      Result_File.l_time[k] = accum;
  }
  return;
}



/*
 * Process queueing time
 * Called by meas.c:: update_meas
 */
process_qtime ()
{
  register oqp    *tmp;
  register Probe  *prb;

  register int    j,k;
  register int    accum;
  register int    nrcvd;

  /* Start with the tail */
  tmp = Result_File.tail_oqp;

  for (k=0; k<Total_Hops; k++) {
    /* Check the hops in order */
    prb = *(Batch_File + k) + current_probe;
    nrcvd = 0;
    accum = 0;

    for (j=0; j<Num_Probes; j++, prb++) {
      /* Look at each entry of the batch file and 
	 compute first order results */
      if (prb->rcvflag) {
	nrcvd ++;
	accum += (prb->T4 - prb->T1); 
      }
    }
    /* Calculate mean */
    /* Keep the mean of cumulative queueing times
       Separate them only during display 
       */
    if (nrcvd)
      tmp->q_time[k] = (int) ((double) accum / (double) nrcvd);
  }
  return;
}



/*
 * Process total symmetric delay
 * Called by meas.c:: update_meas
 */
process_dtime ()
{
  register oqp    *tmp;
  register Probe  *prb;

  register int    j,k;
  register int    accum, accum2;
  register int    nrcvd;

  /* Start with the tail */
  tmp = Result_File.tail_oqp;

  accum2 = 0;
  for (k=0; k<Total_Hops; k++) {
    /* Check the hops in order */
    prb = *(Batch_File + k) + current_probe;
    nrcvd = 0;
    accum = 0;

    for (j=0; j<Num_Probes; j++, prb++) {
      /* Look at each entry of the batch file and 
	 compute first order results */
      if (prb->rcvflag) {
	nrcvd ++;
	accum += (prb->T4 - prb->T1); 
      }
    }
    /* Calculate mean */
    if (nrcvd) {
      tmp->d_time[k] = (int) ((double) accum / (double) nrcvd) - accum2;
      /*      accum2 += tmp->d_time[k];
	      tmp->d_time[k] /= 2;
	      */
    }
  }
  return;
}



/*
 * Add a node to the list of results 
 * Initialize the next node
 * Called by: meas.c:: update_batch
 */
add_results_node ()
{
  register oqp *tmp;
  register int i;

  Result_File.num_batches ++;
  tmp = (oqp *) malloc (sizeof (oqp));
  (Result_File.tail_oqp)->next = tmp;
  Result_File.tail_oqp = tmp;

  /* malloc internal struct of oqp */
  tmp->o_time = (int *) malloc (sizeof (int) * Total_Hops);
  tmp->q_time = (int *) malloc (sizeof (int) * Total_Hops);
  tmp->d_time = (int *) malloc (sizeof (int) * Total_Hops);
  tmp->p_loss = (double *) malloc (sizeof (double) * Total_Hops);
  tmp->next = NULL;
  for (i=0; i<Total_Hops; i++) {
    tmp->o_time[i] = 0;
    tmp->q_time[i] = 0;
    tmp->d_time[i] = 0;
    tmp->p_loss[i] = (double) 0;
  }
  return;
}


/*
 * Kill the measurement process
 * Display (ASCII) results
 * Called by: probe.c interrupt
 */
void
meas_kill (int dummy)
{
  register int i;

  printf ("\nEnding Measurements.... Total Batches: %d\n", batch_count);
  for (i=0; i<Result_File.num_batches; i++)
    display_results (i);
#ifdef GUI_ENABLE
  gui_exit ();
#endif
  exit (0);
}
